# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021, Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------
from threading import Lock


class Singleton(type):
    """
    This is a thread-safe implementation of Singleton pattern.
    ---------
    @author:    Hieu Pham.
    @created:   24-05-2021.
    @modified:  24-05-2021.
    """

    # We will keep created instance.
    _instances = {}
    # And lock for thread-safe.
    _lock: Lock = Lock()

    # We now have a lock object that will be used to synchronize threads during
    # first access to the Singleton.
    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.
        with cls._lock:
            # The first thread to acquire the lock, reaches this conditional,
            # goes inside and creates the Singleton instance. Once it leaves the
            # lock block, a thread that might have been waiting for the lock
            # release may then enter this section. But since the Singleton field
            # is already initialized, the thread won't create a new object.
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        # Return result.
        return cls._instances[cls]



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021, Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------
from typing import Any, Union, List
from abc import ABC, abstractmethod
from pydps.serializations import JsonSerializable, JsonSerializer, JsonDeserializer


class BaseLeaf(JsonSerializable, ABC):
    """
    This class is basic element to build tree node.
    ---------
    @author:    Hieu Pham.
    @created:   21-09-2021.
    @updated:   21-09-2021.
    """

    @property
    def is_root(self):
        return self._parent is None

    @property
    def parent(self):
        return self._parent

    @property
    def is_leaf(self):
        return True

    def __init__(self, parent: Any = None, **kwargs):
        """
        Construct new instance.
        :param parent:  parent node of leaf.
        :param kwargs:  additional keyword arguments.
        """
        super().__init__(**kwargs)
        # Validate parent node.
        assert isinstance(parent, BaseLeaf) or parent is None, 'Parent of node must be at least instance of BaseLeaf.'
        # Assign parent node.
        self._parent = parent


class BaseNode(BaseLeaf, ABC):
    """
    This class is basic element to build tree structure.
    ---------
    @author:    Hieu Pham.
    @created:   21-09-2021.
    @updated:   21-09-2021.
    """

    def __init__(self, parent: BaseLeaf = None, **kwargs):
        """
        Construct new instance.
        :param parent:  parent node of node.
        :kwargs:        additional keyword arguments.
        """
        super().__init__(parent=parent, **kwargs)
        self._nodes = list()

    def attach(self, nodes: Union[BaseLeaf, List[BaseLeaf]]):
        """
        Attach given node(s) to tree.
        :param nodes:   given node(s).
        :return:        tree and node(s).
        """
        # Attach single node.
        if isinstance(nodes, BaseLeaf):
            nodes._parent = self
            self._nodes.append(nodes)
        # Attach nodes list.
        elif isinstance(nodes, list):
            for node in nodes:
                self.attach(node)
        # Return result.
        return self, nodes

    def detach(self, indexes: [int, List[int]]):
        """
        Detach node(s) from tree.
        :param indexes: index(es) of detached node(s).
        :return:        tree and node(s).
        """
        node = None
        # Detach single node.
        if isinstance(indexes, int):
            node = self._nodes.pop(indexes)
            node._root = None
        # Detach nodes list.
        elif isinstance(indexes, list):
            node = [self.detach(i)[-1] for i in indexes]
        # Return result.
        return self, node

    def clean(self):
        """
        Detach all nodes.
        :return: tree and node(s).
        """
        return self.detach([i for i in range(len(self._nodes))])

    @abstractmethod
    def __serialize__(self) -> dict:
        """
        Serialize object to dict structures.
        :return:    dict structures.
        """
        data = super().__serialize__()
        data.update({'nodes': [JsonSerializer.encode(node) for node in self._nodes]})
        return data

    @abstractmethod
    def __deserialize__(self, data: dict = {}):
        """
        Deserialize dict structures to object.
        :param data:    dict structures.
        :return:        object itself.
        """
        self.clean()
        self.attach([JsonDeserializer().decode(dat) for dat in data.get('nodes', {})])
        return self



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021, Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------
from typing import Any
from abc import ABC, abstractmethod


class CommandInterface(ABC):
    """
    This is command interface declares a method for executing a command.
    ---------
    @author:    Hieu Pham.
    @created:   24-05-2021.
    @modified:  24-05-2021.
    """

    @abstractmethod
    def do(self) -> Any:
        """
        Perform the command.
        :return: anything.
        """
        pass

    @abstractmethod
    def undo(self) -> Any:
        """
        Undo the command.
        :return: anything.
        """
        pass



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021, Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------
import importlib
from typing import Any
from abc import ABC, abstractmethod
from pydps.creation import Singleton
from json import JSONEncoder, JSONDecoder, JSONDecodeError


class JsonSerializable(ABC):
    """
    This interface class provide json serialization feature to inherited classes.
    ---------
    @author:    Hieu Pham.
    @created:   21-09-2021.
    @updated:   21-09-2021.
    """

    def __init__(self, **kwargs):
        """
        Construct new instance.
        :param kwargs:  additional keyword arguments.
        """
        super().__init__()

    @abstractmethod
    def __serialize__(self, **kwargs) -> dict:
        """
        Serialize json-liked data.
        :return: json-liked data.
        """
        return {'class': self.__class__.__name__, 'module': self.__module__}

    def __deserialize__(self, data: dict = dict):
        """
        Deserialize json-liked data.
        :param data: json-liked data.
        :return:     object itself.
        """
        return self


class JsonSerializer(JSONEncoder, metaclass=Singleton):
    """
    This singleton class will serialize object to json dict structures.
    ---------
    @author:    Hieu Pham.
    @created:   21-09-2021.
    @updated:   21-09-2021.
    """

    def __init__(self):
        """
        Construct new instance.
        """
        super().__init__(default=self.default)

    def default(self, o: Any) -> Any:
        """
        Override default encoding handle.
        :param o:   object to be serialized.
        :return:    any.
        """
        return o.__serialize__() if isinstance(o, JsonSerializable) else super().default(o)


class JsonDeserializer(JSONDecoder, metaclass=Singleton):
    """
    This singleton class will deserialize json dict structures to object.
    ---------
    @author:    Hieu Pham.
    @created:   21-09-2021.
    @updated:   21-09-2021.
    """

    def __init__(self):
        """
        Construct new instance.
        """
        super().__init__(object_hook=self.object_hook)

    def object_hook(self, data: dict = None) -> Any:
        """
        Override parent object hook method.
        :param data:    structures dict.
        :return:        any.
        """
        # Validate data.
        if 'class' in data and 'module' in data:
            # Dynamic load module and class by name.
            class_name, module_name = data.pop('class'), data.pop('module')
            classes = getattr(importlib.import_module(module_name), class_name)
            # Create instance and then return.
            if issubclass(classes, JsonSerializable):
                return classes(**data).__deserialize__(data)
        # Otherwise, raise an error.
        raise JSONDecodeError("Cannot deserialize invalid data format.")
